#include <windows.h>
#include <atlbase.h>
#include <atlconv.h>
#include <comdef.h>
#include <stdio.h>

#include "MJInputSource.h"
#include "Base64Coder.h"
#include "CharacterHelper.h"
#include "SmartPtr.h"

#include "vorbis\vorbisfile.h"

#include "vcedit.h"
#include "VorbisComments.h"

HINSTANCE g_hDllInstance = NULL;

// TO DO -- should probably support these (need to adjust gain by 6 dB though)
//REPLAYGAIN_TRACK_GAIN
//REPLAYGAIN_TRACK_PEAK

#define NUMBER_FIELDS	13
LPCWSTR g_aryMJFields[NUMBER_FIELDS] =		{ MF_NAME,	MF_ARTIST,	MF_ALBUM,	MF_GENRE,	MF_NOTES,	MF_LYRICS,	_T("MJMETADATA"),	MF_YEAR,	MF_TRACKNUMBER,	MF_COMMENT,	MF_COMMENT,		MF_COMPOSER,	MF_KEYWORDS		};
LPCWSTR g_aryOGGFields[NUMBER_FIELDS] =		{ L"TITLE",	L"ARTIST",	L"ALBUM",	L"GENRE",	L"NOTES",	L"LYRICS",	L"MJMETADATA",		L"DATE",	L"TRACKNUMBER",	L"COMMENT",	L"DESCRIPTION",	L"COMPOSER",	L"KEYWORDS"		};
BOOL g_aryOGGFieldNumeric[NUMBER_FIELDS] =	{ FALSE,	FALSE,		FALSE,		FALSE,		FALSE,		FALSE,		FALSE,				TRUE,		TRUE,			FALSE,		FALSE,			FALSE,			FALSE			};

class OggFileInfo : public CMJFileInfo 
{
public:
	
	OggFileInfo();
	~OggFileInfo();

	BOOL GetAttribute(LPCTSTR Name, LPTSTR Value, int * nCharacters);
	BOOL SetAttribute(LPCTSTR Name, LPCTSTR Value);
	BOOL GetFormatString(LPTSTR Value, int * nCharacters);

	int DoCommand(MJ_FILE_INFO_COMMANDS nCommandID, int nParam1 = 0, int nParam2 = 0);

	BOOL Open(LPCTSTR lpszFilename);
	BOOL Close();
	
private:	

	// for bitrate and duration
	int m_nBitrate;
	int m_nDuration;
	int m_nSampleRate;
	int m_nChannels;

	TCHAR m_cFilename[MAX_PATH];

	// for comments
	FILE * m_pFile;
	VorbisComments m_Comments;
	vcedit_state * m_pState;

	BOOL m_bFirstSet;
	BOOL m_bChanged;
	BOOL m_bRemoveTag;
};

OggFileInfo::OggFileInfo()
{
	m_bChanged = FALSE;
	m_bRemoveTag = FALSE;
	m_bFirstSet = TRUE;

	m_pFile = NULL;
	m_nBitrate = 0;
	m_nDuration = 0;
	m_nSampleRate = 0;
	m_nChannels = 0;	

	m_pState = new vcedit_state();
	memset(m_pState, 0, sizeof(*m_pState));
}

OggFileInfo::~OggFileInfo()
{
	Close();	
}

BOOL OggFileInfo::Open(LPCTSTR lpszFilename)
{	
	BOOL bRetVal = FALSE;

	m_bFirstSet = TRUE;
	m_bChanged = FALSE;
	m_bRemoveTag = FALSE;

	if ((lpszFilename != NULL) && (lpszFilename[0] != 0))
	{
		_tcscpy(m_cFilename, lpszFilename);

		// get header info
		m_pFile = _tfopen(m_cFilename, _T("rb"));
		if (m_pFile != NULL)
		{
			OggVorbis_File * pOggFile = new OggVorbis_File;
			if (ov_open(m_pFile, pOggFile, NULL, 0) >= 0)
			{
				m_nBitrate = ov_bitrate(pOggFile, -1);	
				m_nDuration = (int) ov_time_total(pOggFile, -1);

				vorbis_info * pVorbisInfo = ov_info(pOggFile, -1);
				if (pVorbisInfo)
				{
					m_nSampleRate = pVorbisInfo->rate;
					m_nChannels = pVorbisInfo->channels;
				}

				ov_clear(pOggFile);
				delete pOggFile;

				// get comment info
				fclose(m_pFile); m_pFile = NULL;
                m_pFile = _tfopen(m_cFilename, _T("rb"));
				if (m_pFile != NULL)
				{
					if (vcedit_open(m_pState, m_pFile) == 0)
					{
						m_Comments.retrieve(vcedit_comments(m_pState));
						bRetVal = TRUE;
					}
				}
			}
		}
	}
	
	return bRetVal;
}

BOOL OggFileInfo::Close()
{
	if (m_bChanged)
	{
		m_bChanged = FALSE; // have to do this to prevent this code from being executed twice
		LPTSTR filepart;
		
		// calculate a temporary file name
		TCHAR fullpath[MAX_PATH];
		GetFullPathName(m_cFilename, MAX_PATH, fullpath, &filepart);
		*filepart = 0;
		
		TCHAR tempfile[MAX_PATH];
		GetTempFileName(fullpath, _T("vce"), 0, tempfile);
		lstrcpy(fullpath, m_cFilename);
		
		// Write the new vorbis file
		FILE *output = _tfopen(tempfile, _T("wb"));
		
		if (m_bRemoveTag) m_Comments.removeAll();
		m_Comments.store(vcedit_comments(m_pState));
					
		if (vcedit_write(m_pState, output) == -1)
		{
			fclose(m_pFile);
			m_pFile = NULL;
			fclose(output);
			DeleteFile(tempfile);
			return FALSE;
		}
		
		// close the files
		fclose(m_pFile);
		m_pFile = NULL;
		fclose(output);
		
		// rename the new vorbis file
		if (DeleteFile(fullpath) == 0)
			DeleteFile(tempfile);
		
		MoveFile(tempfile, fullpath);
	}

	if (m_pFile)
	{
		fclose(m_pFile);
		m_pFile = NULL;
	}

	if (m_pState)
	{
		vcedit_clear(m_pState);
		m_pState = NULL;
	}

	return TRUE;
}

BOOL OggFileInfo::GetAttribute(LPCTSTR Name, LPTSTR Value, int * nCharacters)
{
	// clear
	_tcscpy(Value, _T(""));

	// handle "special" fields
	if (_tcscmp(Name, MF_BITRATE) == 0)
	{
		if (*nCharacters < 16) { *nCharacters = 16; return FALSE; }
		_stprintf(Value, _T("%d"), m_nBitrate / 1000);
		return TRUE;
	}	
	else if (_tcscmp(Name, MF_DURATION) == 0)
	{
		if (*nCharacters < 16) { *nCharacters = 16; return FALSE; }
		_stprintf(Value, _T("%d"), m_nDuration);
		return TRUE;
	}	
	else if (_tcscmp(Name, MF_IMAGEFILE) == 0)
	{
		LPSTR temp = m_Comments.getValue("COVERARTMIME");
		if (temp == NULL) return FALSE;
		
		TCHAR cExt[6];
		if (strstr(temp, "image/bmp") != NULL)
			_tcscpy(cExt, _T("bmp"));
		else if (strstr(temp, "image/jpeg") >= 0)
			_tcscpy(cExt, _T("jpg"));
		else if (strstr(temp, "image/jpg") >= 0)
			_tcscpy(cExt, _T("jpg"));
		else if (strstr(temp, "image/gif") >= 0)
			_tcscpy(cExt, _T("gif"));
		else if (strstr(temp, "image/png") >= 0)
			_tcscpy(cExt, _T("png"));
		else if (strstr(temp, "image/tiff") >= 0)
			_tcscpy(cExt, _T("tif"));
		else if (strstr(temp, "image/tif") >= 0)
			_tcscpy(cExt, _T("tif"));

		TCHAR cPath[MAX_PATH]; ::GetTempPath(MAX_PATH, cPath);
		TCHAR cFilename[MAX_PATH]; _stprintf(cFilename, _T("%spic.%s"), cPath, cExt);

		if (int(_tcslen(cFilename)) > *nCharacters) { *nCharacters = _tcslen(cFilename); return FALSE; }

		_tcscpy(Value, cFilename);
		
		LPSTR pPicture = m_Comments.getValue("COVERART");
		if (pPicture == NULL) return FALSE;

		int nBytes = strlen(pPicture);

		Base64Coder Coder;
		Coder.Decode((const PBYTE) pPicture, nBytes);
		LPCSTR pDecoded = Coder.DecodedMessage();

		int nCurrentPosition = 0;
		FILE * hFile = _tfopen(cFilename, _T("wb"));
		while (hFile)
		{
			size_t nWritten = fwrite(&pDecoded[nCurrentPosition], 1, nBytes, hFile);
			nBytes = nBytes - nWritten;
			nCurrentPosition += nWritten;
			
			if ((nWritten == 0) || (nBytes <= 0))
			{
				fclose(hFile);
				hFile = NULL;
			}
		}		

		return TRUE;
	}	

	// handle "regular" fields
	for (int z = 0; z < (NUMBER_FIELDS + 1); z++)
	{
		if ((z == NUMBER_FIELDS) || (_tcscmp(Name, g_aryMJFields[z]) == 0))
		{
			// get the field name
			TCHAR cField[256] = {0};
			if (z >= NUMBER_FIELDS)
			{
				_tcscpy(cField, _T("MediaJukebox:")); 
				_tcscat(cField, Name); 
				_tcsupr(cField);
			}
			else
			{
				_tcscpy(cField, g_aryOGGFields[z]);
			}

			// get the value from the tag
			CSmartPtr<char> spFieldUTF8((char *) GetUTF8FromUTF16(cField), TRUE);
			CSmartPtr<WCHAR> spValue(GetUTF16FromUTF8((UCHAR *) m_Comments.getValue(spFieldUTF8.GetPtr())), TRUE);
			if (spValue != NULL)
			{
				if ((wcslen(spValue) + 1) > *nCharacters) 
				{ 
					*nCharacters = wcslen(spValue) + 1; 
					return FALSE; 
				}
				
				_tcscpy(Value, spValue);
				return TRUE;
			}
		}
	}
	
	// didn't find the value -- so return failure
	return FALSE;
}

BOOL OggFileInfo::SetAttribute(LPCTSTR Name, LPCTSTR Value)
{
	m_bChanged = TRUE;

	// remove all MJ-specific info on the first set
	if (m_bFirstSet)
	{
		m_bFirstSet = FALSE;

		for (int z = 0; z < m_Comments.getCount(); z++)
		{
			if (strstr(m_Comments.getName(z), "MEDIAJUKEBOX:") != NULL)
				m_Comments.removeTag(z--);
		}

		m_Comments.removeTag("MJMETADATA");
	}

	// handle "regular" fields
	BOOL bSet = FALSE;
	for (int z = 0; z < (NUMBER_FIELDS + 1); z++)
	{
		if ((z == NUMBER_FIELDS) || (_tcscmp(Name, g_aryMJFields[z]) == 0))
		{
			// get the field name
			TCHAR cField[256] = {0};
			if (z >= NUMBER_FIELDS)
			{
				_tcscpy(cField, _T("MediaJukebox:")); 
				_tcscat(cField, Name); 
				_tcsupr(cField);
			}
			else
			{
				_tcscpy(cField, g_aryOGGFields[z]);
			}

			// always remove the tag (in case a field maps to multiple OGG keywords (i.e. MF_COMMENT))
			CSmartPtr<char> spFieldUTF8((char *) GetUTF8FromUTF16(cField), TRUE);
			m_Comments.removeTag(spFieldUTF8);
			
			// set (only the first time)
			BOOL bNumeric = (z < NUMBER_FIELDS) ? g_aryOGGFieldNumeric[z] : FALSE;
			if ((bSet == FALSE) && (bNumeric ? (_ttoi(Value) != 0) : (_tcslen(Value) != 0)))
			{
				CSmartPtr<char> spValueUTF8((char *) GetUTF8FromUTF16(Value), TRUE);
				m_Comments.addTag(spFieldUTF8, spValueUTF8);
				bSet = TRUE;
			}
		}
	}

	// special fields
	if (_tcscmp(Name, MF_IMAGEFILE) == 0)
	{
		m_Comments.removeTag("COVERARTMIME");
		m_Comments.removeTag("COVERART");

		if (_tcslen(Value) != 0)
		{
			USES_CONVERSION;

			TCHAR drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
			_tsplitpath( Value, drive, dir, fname, ext );

			char MimeType[64] = "image/"; 
			strcat(MimeType, (_tcsicmp(ext, _T(".jpg")) == 0) ? "jpeg" : T2A(&ext[1]));

			if (strlen(MimeType) != 0)
				m_Comments.addTag("COVERARTMIME", MimeType);

			CSmartPtr<BYTE> spPicture; fpos_t pos = 0;
			FILE * hFile = _tfopen(Value, _T("rb"));
			if (hFile)
			{
				fseek(hFile, 0, SEEK_END);
				fgetpos(hFile, &pos);
				fseek(hFile, 0, SEEK_SET);
				
				spPicture.Assign(new BYTE [pos], TRUE);

				int CurPos = 0, Len = pos;
				while (hFile)
				{
					size_t Written = fread(&spPicture[CurPos], 1, Len, hFile);
					Len = Len - Written;
					CurPos += Written;
					
					if (Written == 0 || Len <= 0)
					{
						fclose(hFile);
						hFile = NULL;
					}
				}
			}		

			if (spPicture != NULL)
			{
				Base64Coder coder;
				coder.Encode((const PBYTE) spPicture, pos);
				LPCSTR str = coder.EncodedMessage();

				if (strlen(str) != 0)
					m_Comments.addTag("COVERART", str);
			}
		}
	}
	
	return TRUE;
}

BOOL OggFileInfo::GetFormatString(LPTSTR Value, int * nCharacters)
{
	if (*nCharacters < 512) { *nCharacters = 512; return FALSE; }
	
	TCHAR szTemp[512];

	_stprintf(szTemp, _T("Bitrate: %d Kbps\r\n"), m_nBitrate / 1000);
	_tcscpy(Value, szTemp);
		
	_stprintf(szTemp, _T("Frequency: %d Hz\r\n"), m_nSampleRate);
	_tcscat(Value, szTemp);
		
	_stprintf(szTemp, _T("Channels: %d\r\n"), m_nChannels);
	_tcscat(Value, szTemp);
	
	int nTagBytes = 0;
	for (int z = 0; z < m_Comments.getCount(); z++)
		nTagBytes += strlen(m_Comments.getValue(z));

	_stprintf(szTemp, _T("Tag: (%d fields -- %d data bytes)\r\n"), m_Comments.getCount(), nTagBytes);
	_tcscat(Value, szTemp);

	for (int z = 0; z < m_Comments.getCount(); z++)
	{
		CSmartPtr<WCHAR> spName(GetUTF16FromUTF8((UCHAR *) m_Comments.getName(z)), TRUE);
		CSmartPtr<WCHAR> spValue(GetUTF16FromUTF8((UCHAR *) m_Comments.getValue(z)), TRUE);

		_stprintf(szTemp, _T("   %s: %s\r\n"), spName.GetPtr(), (wcslen(spValue) < 128) ? spValue.GetPtr() : _T("<too large to display>"));
		_tcscat(Value, szTemp);
	}

	return TRUE;
}

int OggFileInfo::DoCommand(MJ_FILE_INFO_COMMANDS nCommandID, int nParam1, int nParam2)
{
	if (nCommandID == MJ_FILE_INFO_COMMAND_REMOVE_TAG)
	{
		m_bChanged = TRUE; // have to do this to prevent this code from being executed twice
		m_bRemoveTag = TRUE;

		return 0;
	}
	else if (nCommandID == MJ_FILE_INFO_COMMAND_GET_NO_FIELD_TRANSLATION)
	{
		return TRUE;
	}

	return -1;
}

extern "C"
{
	__declspec (dllexport) int GetFileInfo(HINSTANCE hDllInstance, CMJFileInfo **  ppFileInfo)
	{
		g_hDllInstance = hDllInstance;
		*ppFileInfo = (CMJFileInfo *) new OggFileInfo;
		return 1;
	}
}


